local CowTip = CowTip
local CowTip_LuaText = CowTip:NewModule("LuaText", "LibRockEvent-1.0", "LibRockHook-1.0", "LibRockTimer-1.0")
local self = CowTip_LuaText

--[===[@debug@
-- function is not needed at all when debug is off
local function debugTableValues(table, tabs, stop)
	if not tabs then tabs = 0 end
	if stop then
		for k,v in pairs(table) do
			local line = ("  "):rep(tabs)
			line = line .. ("[%s]"):format(tonumber(k) or tostring(k))
			line = line .. (" = %s,"):format(tonumber(v) or tostring(v))
			ChatFrame4:AddMessage(line)
		end
	else
		for k,v in pairs(table) do
			local line = ("  "):rep(tabs)
			line = line .. ("[%s] = "):format(tonumber(k) or tostring(k))

			if type(v) == 'table' then
				ChatFrame4:AddMessage(line .. "{")
				debugTableValues(v, tabs + 1, true)
				ChatFrame4:AddMessage(("  "):rep(tabs) .. "}")
			else
				ChatFrame4:AddMessage("  " .. line .. tostring(v) .. ",")
			end
		end

	end
end
--@end-debug@]===]

local function debug(...)
	--[===[@debug@
	local first = select(1,...)
	if type(first) == 'table' then
		ChatFrame4:AddMessage("{")
		debugTableValues(first)
		ChatFrame4:AddMessage("}")
	else
		local text = ""
		for i = 1, select('#', ...) do
			text = text .. tostring(select(i, ...))
		end
		ChatFrame4:AddMessage(text)
	end
	--@end-debug@]===]
end

--[[
-- Text-provider-code
--]]
local funcs = {}

local ScriptEnv
function CowTip_LuaText:SetScriptEnv(arg)
--	CowTip_LuaText.ScriptEnv = arg
	ScriptEnv = arg
end

local mouseoverUnit = CowTip:GetMouseoverUnit()

function CowTip_LuaText:NewText(code)
	mouseoverUnit = CowTip:GetMouseoverUnit()
	local func
	local lua_string = 'return function(unit) if not unit then return \"\" end; '..code..' end'
	local create_func, err = loadstring(lua_string)
	if create_func then
		func = create_func()
		setfenv(func, ScriptEnv)
	else
		geterrorhandler()(err)
	end
	if not func then func = function() return "" end end
	funcs[code] = func
end

function CowTip_LuaText:RemoveText(code)
--	texts[code] = nil
	funcs[code] = nil
end

function CowTip_LuaText:GetText(code, unit)
--	return texts[code]
	return string.format(funcs[code](unit))
end

--[[
--
--]]

function CowTip_LuaText:OnProfileEnable()
	wipe(funcs)
	for k,v in pairs(self.db.profile.lefts) do
		self:NewText(v, {})
	end
	for k,v in pairs(self.db.profile.rights) do
		self:NewText(v, {})
	end
end

local afk_list
local dnd_list
local offline_list


local L = CowTip:L("CowTip_Text")

local leftGameTooltipStrings = CowTip.leftGameTooltipStrings
local rightGameTooltipStrings = CowTip.rightGameTooltipStrings

local _G = _G
local IsShiftKeyDown = _G.IsShiftKeyDown
local ChatFontNormal = _G.ChatFontNormal
local CreateFrame = _G.CreateFrame
local UIParent = _G.UIParent
local GameTooltip = _G.GameTooltip
local ipairs = _G.ipairs
local GetNumFactions = _G.GetNumFactions
local GetFactionInfo = _G.GetFactionInfo
local UnitPlayerControlled = _G.UnitPlayerControlled
local UnitIsPlayer = _G.UnitIsPlayer
local UnitFactionGroup = _G.UnitFactionGroup
local UnitIsConnected = _G.UnitIsConnected
local UnitIsVisible = _G.UnitIsVisible
local UnitIsPVP = _G.UnitIsPVP
local UnitExists = _G.UnitExists

local kwargs = { unit = 'mouseover' }

local opt
local function makeOptions()
	if not opt then
		opt = {
			name = L["Lua Line texts"],
			desc = L["Here you can describe the text on the assorted lines of the tooltip in Lua."],
			type = 'group',
			args = {
				linetext = {
					type = 'execute',
					name = "Edit",
					desc = "Edit Line Texts",
					func = self.GUI,
					passValue = self,
				}
			},
		}
	--[[
		local function getLeft(i)
			return self.db.profile.lefts[i] or ""
		end
		local function getRight(i)
			return self.db.profile.rights[i] or ""
		end
		local function setText(side, i, value)
			local old = self.db.profile[side][i]
			self:NewText(value)
			self.db.profile[side][i] = value
			if old then
				self:RemoveText(old)
			end
		end
		
		local function setLeft(i, value)
			setText("lefts", i, value)
		end
		local function setRight(i, value)
			setText("rights", i, value)
		end
		
		for i = 1,30 do
			local line_opt = {
				name = L["Line %d"]:format(i),
				desc = "",
				type = 'group',
				groupType = 'inline',
				args = {
					left = {
						name = L["Left"],
						desc = L["Left"],
						type = 'string',
						multiline = true,
						get = getLeft,
						set = setLeft,
						passValue = i,
						usage = "<code>",
					},
					right = {
						name = L["Right"],
						desc = L["Right"],
						type = 'string',
						multiline = true,
						get = getRight,
						set = setRight,
						passValue = i,
						usage = "<code>",
					},
				},
				order = i,
			}
			opt.args["line" .. i] = line_opt
		end -- for i = 1,9--]]
		self:RegisterCowTipOption(opt)
	end -- if not opt
end

function CowTip_LuaText:OnInitialize()
	self.db = CowTip:GetDatabaseNamespace("LuaText")
	CowTip:SetDatabaseNamespaceDefaults("LuaText", "profile", {
		lefts = {
			['*'] = "",
			[1] = [[local name = UnitPVPName(unit) or Name(unit)
if not name then
  return UNKNOWN
end
local _, class = UnitClass(unit)
local color = "%1"
if UnitIsPlayer(unit) then
  color = Colorize("%1", ClassColor(unit))
else
  color = Colorize("%1", HostileColor(unit))
end
name = name:gsub(Name(unit), color)
return name
]],
			[2] = [[local guild, rankname, rankindex = GetGuildInfo(unit)
if not guild then
	guild = NPCGuild()
	if not guild then return '' end
	return "<%s>", guild
end
local pguild, _, prankindex = GetGuildInfo("player")
if pguild == guild then
  guild = ("|cff00cc00<%s>|r"):format(guild)
  local rankcolor
  if rankindex > prankindex then
    rankcolor = "|cffff0000%s|r"
  elseif rankindex < prankindex then
    rankcolor = "|cff00ff00%s|r"
  else
    rankcolor = "|cffffff00%s|r"
  end
  rankname = rankcolor:format(rankname)
else
  guild = ("<%s>"):format(guild)
end
return "%s (%d) of %s", rankname, rankindex, guild
]],
			[3] = [[return Faction() or '']],
			[4] = ("return \"%s\""):format(TARGET),
			[5] = [[local level, race = Level(unit), SmartRace(unit)
local class = ""
if UnitIsPlayer(unit) then
  class = Colorize(Class(unit), ClassColor(unit))
else
  class = ""
end
level = Colorize(level, DifficultyColor(unit))
local diff = ShortClassification(unit)
return "Level %s%s %s %s", level, diff, race, class
]],
			[6] = "return 'Status:'",
			[7] = "return 'Zone:'",
			[8] = "return '|T%s:24|t %s', TalentSpecIcon(unit) or '', TalentSpecName(unit) or ''",
--			"[if [IsPlayer or [IsEnemy and not IsPet]] then ClassColor][NameRealm]", -- [1]
--[[			L["Target:"], -- [2]
			L["Guild:"], -- [3]
			L["Guild rank:"], -- [4]
			L["Level:"], -- [5]
			L["Race:"], -- [6]
			L["Class:"], -- [7]
			L["Faction:"], -- [8]
			L["Status:"], -- [9]
			L["Health:"], -- [10]
			"[if HasMP then TypePower ':']", -- [11]
			L["Experience:"], -- [12]
			L["Threat:"], -- [13]
			L["Happiness:"], -- [14]
			L["Zone:"], -- [15]
			(_G.TALENTS or "Talents") .. ":", -- [16]--]]
		},
		rights = {
			['*'] = "",
			[1] = "",
			[2] = "",
			[3] = "",
			[4] = [[local name = Name(unit .. "target")
if not name then
	return NONE
elseif name == UnitName("player") then
	return Colorize(("<<%s>>"):format(YOU), 0, 0, 255)
else
	return Colorize(name, SmartClassColor(unit .. "target"))
end
]],
			[5] = "",
			[6] = [[return Offline(unit) or AFK(unit) or DND(unit) or Status(unit) or '']],
			[7] = "return Zone() or ''",
			[8] = "return TalentSpec(unit) or ''",
	--[[		"", -- [1]
			"[if Target:IsPlayer or (Target:IsEnemy and not Target:IsPet) then ClassColor(unit=Target)][if IsUnit('player', Target) then '<<" .. L["You"] .. ">>' else Target:NameRealm or '" .. L["None"] .. "']", -- [2]
			"[if Guild and Guild = 'player':Guild then Green][Guild:Angle]", -- [3]
			"[GuildRank]", -- [4]
			"[Classification] [Level]", -- [5]
			"[if IsPlayer then Race else Creature]", -- [6]
			"[if IsPlayer or [IsEnemy and not IsPet] then Class:ClassColor]", -- [7]
			"[Faction]", -- [8]
			"[Status]", -- [9]
			"[FractionalHP:Short] [PercentHP:Percent:Paren]", -- [10]
			"[if HasMP then FractionalMP:Short ' ' PercentMP:Percent:Paren]", -- [11]
			"[if XP ~= 0 then FractionalXP:Short ' ' PercentXP:Percent:Paren]", -- [12]
			"[if HasThreat then FractionalThreat:Short ' ' PercentThreat:Percent:Paren]", -- [13]
			"[if IsUnit('pet') then HappyText]", -- [14]
			"[Zone]", -- [15]
			"[if TalentSpec then TalentTree ' ' TalentSpec:Paren]", -- [16]--]]
		}
	})
	makeOptions()
end

function CowTip_LuaText:OnEnable()
	CowTip:GetModule("Text"):ToggleActive(false)
--	ScriptEnv = CowTip_LuaText.ScriptEnv
	for k,v in pairs(self.db.profile.lefts) do
		self:NewText(v, {})
	end
	for k,v in pairs(self.db.profile.rights) do
		self:NewText(v, {})
	end
	
	self:AddEventListener("UPDATE_FACTION")
	
	self:AddHook(GameTooltip, "GetUnit", "GameTooltip_GetUnit")
	self:UPDATE_FACTION()
	-- gui	
end

local function char_count(str, char) 
    if not str then
        return 0
    end

    local count = 0 
    local byte_char = string.byte(char)
    for i = 1, #str do
        if string.byte(str, i) == byte_char then
            count = count + 1 
        end 
    end 
    return count
end


local AceGUI = LibStub("AceGUI-3.0")
local win
function CowTip_LuaText:GUI()
	if not win then
		win = AceGUI:Create("Frame")
		win:SetWidth(800)
		local frame = AceGUI:Create("ScrollFrame")
		frame:SetLayout("flow")
		frame.width = "fill"
		frame.height = "fill"
		for i = 1,30 do
			local grp = AceGUI:Create("InlineGroup")
			grp.width = "fill"
			grp:SetLayout("flow")
			grp:SetTitle("Line " .. i)
			frame:AddChild(grp)

			local boxcallback = function(widget)
					local lines = char_count(widget:GetText(), '\n') + 4
					widget:SetNumLines(lines)
					-- FIXME this is a hack
	--				grp:SetHeight(80 + lines*14)
					local height = win.frame:GetHeight()
					if math.floor(height)%2 == 0 then
						win:SetHeight(height + 1)
					else
						win:SetHeight(height - 1)
					end
				end

			local text1 = AceGUI:Create("MultiLineEditBox")
			text1:SetLabel("Left")
			text1:SetWidth(350)
			text1:SetText(self.db.profile.lefts[i]:gsub("\|", "\|\|"))
			local lines = char_count(self.db.profile.lefts[i],'\n') + 4
			text1:SetNumLines((lines < 4 and 4) or lines)
			text1:SetCallback("OnTextChanged", boxcallback)
			text1:SetCallback("OnEnterPressed", function(widget)
					local old = self.db.profile.lefts[i]
					local new = widget:GetText():gsub("\|\|","\|")
					if old == new then return end
					self:NewText(new)
					self:RemoveText(old)
					self.db.profile.lefts[i] = new
				end)
		
			grp:AddChild(text1)
		
		
			local text2 = AceGUI:Create("MultiLineEditBox")
			text2:SetLabel("Right")
			text2:SetWidth(350)
			text2:SetText(self.db.profile.rights[i])
			lines = char_count(self.db.profile.rights[i],'\n')
			text2:SetNumLines((lines < 4 and 4) or lines)
			text2:SetCallback("OnTextChanged", boxcallback)
			text2:SetCallback("OnEnterPressed", function(widget)
					self.db.profile.rights[i] = widget:GetText():gsub("\|\|","\|")
				end)
			text2:SetCallback("OnEscapePressed", function(widget)
					widget:SetText(self.db.profile.rights[i]:gsub("\|","\|\|"))
				end)
			grp:AddChild(text2)
		end
		win:AddChild(frame)
	end
	win:Show()
end

function CowTip_LuaText:OnDisable()
	wipe(funcs)
end

-- patterns
local LEVEL_start = "^" .. LEVEL
local PVP = _G.PVP
local factionList = {}

local function scanTT(unit)
	local npcguild, faction, zone
	-- scan for faction
	if not unit then
		wipe(ScriptEnv.orig)
		return
	end
	
	local left_2 = leftGameTooltipStrings[2]:GetText()
	local hasGuild = not left_2:find(LEVEL_start)
	if hasGuild and not UnitIsPlayer(unit) then
		debug("npcguild found ", left_2)
		npcguild = left_2
	end
	local left_3 = leftGameTooltipStrings[3]:GetText()
	local left_4 = leftGameTooltipStrings[4]:GetText()
	local left_5 = leftGameTooltipStrings[5]:GetText()
	local faction
	if not UnitPlayerControlled(unit) and not UnitIsPlayer(unit) then
		faction = hasGuild and left_4 or left_3
		if faction == PVP or not factionList[faction] then
			faction = nil
		end
	end
	-- scan for zone
	local zone
	if not UnitIsVisible(unit) and UnitIsConnected(unit) then
		if hasGuild and faction then
			zone = left_5
		elseif hasGuild or faction then
			zone = left_4
		else
			zone = left_3
		end
	end
	wipe(ScriptEnv.orig)
	ScriptEnv.orig.faction = faction
	ScriptEnv.orig.zone = zone
	ScriptEnv.orig.npcguild = npcguild
end

function CowTip_LuaText:UPDATE_FACTION()
	for i = 1, GetNumFactions() do
		local name = GetFactionInfo(i)
		factionList[name] = true
	end
end

--[[
How normal tooltips look:
{Name}
{[Guild] : ~~~}
Level {Num} {[(Classification)]}
{NPC? [Faction] : ~~~}
{Offline? Offline : !IsVisible? [Zone] : IsPVP? [PvP] : ~~~}
]]

local lastNumRegistered = 0
local linesToAdd = {}
local linesToAddR = {}
local linesToAddG = {}
local linesToAddB = {}
local linesToAddRight = {}
local linesToAddRightR = {}
local linesToAddRightG = {}
local linesToAddRightB = {}

local lastUnitName, lastUnit

function CowTip_LuaText:RefixLines(lastLine)
	lastNumRegistered = 0
	local LinesAdded = 0

	for i = lastLine, GameTooltip:NumLines() do
		local left = leftGameTooltipStrings[i]
		local j = i - lastLine + 1
		linesToAdd[j] = left:GetText()
		local r, g, b = left:GetTextColor()
		linesToAddR[j] = r
		linesToAddG[j] = g
		linesToAddB[j] = b
		local right = rightGameTooltipStrings[i]
		if right:IsShown() then
			linesToAddRight[j] = right:GetText()
			local r, g, b = right:GetTextColor()
			linesToAddRightR[j] = r
			linesToAddRightG[j] = g
			linesToAddRightB[j] = b
		end
		LinesAdded = LinesAdded + 1
	end
	GameTooltip:ClearLines()
	local db = self.db.profile
	local lefts, rights = db.lefts, db.rights
	local num = 0
	local unit = CowTip:GetMouseoverUnit()
	for i = 1, 30 do
		local left, right = lefts[i], rights[i]
		local left_ev, right_ev
		if left and left:trim() == "" then
			left = nil
		else
			--left_ev = tostring(DogTag:Evaluate(left, "Unit", kwargs) or '')
			left_ev = tostring(self:GetText(left, unit))
			if left_ev:trim() == "" then
				left_ev = nil
			end
		end
		if right and right:trim() == "" then
			right = nil
		else
--			right_ev = tostring(DogTag:Evaluate(right, "Unit", kwargs) or '')
			right_ev = tostring(self:GetText(right, unit))
			if right_ev:trim() == "" then
				right_ev = nil
			end
		end
		if right then
			if right_ev or (left and left_ev and left:find("%[")) then
				num = num + 1
				GameTooltip:AddDoubleLine(left_ev or ' ', right_ev, 1, 1, 1, 1, 1, 1)
				local leftFS = leftGameTooltipStrings[num]
				local rightFS = rightGameTooltipStrings[num]
--				DogTag:AddFontString(leftFS, GameTooltip, left or ' ', "Unit", kwargs)
--				DogTag:AddFontString(rightFS, GameTooltip, right, "Unit", kwargs)
			end
		elseif left then
			if left_ev then
				num = num + 1
				GameTooltip:AddLine(left_ev, 1, 1, 1, true)
				local leftFS = leftGameTooltipStrings[num]
				leftFS:SetTextColor(1, 1, 1)
--				DogTag:AddFontString(leftFS, GameTooltip, left, "Unit", kwargs)
			end
		end
	end
	if lastNumRegistered < num then
		lastNumRegistered = num
	end

	for i = 1, LinesAdded do
		local left = linesToAdd[i]
		local right = linesToAddRight[i]
		if right then
			GameTooltip:AddDoubleLine(left, right, linesToAddR[i], linesToAddG[i], linesToAddB[i], linesToAddRightR[i], linesToAddRightG[i], linesToAddRightB[i])
		else
			GameTooltip:AddLine(left, linesToAddR[i], linesToAddG[i], linesToAddB[i], true)
		end
		linesToAdd[i] = nil
		linesToAddR[i] = nil
		linesToAddG[i] = nil
		linesToAddB[i] = nil
		linesToAddRight[i] = nil
		linesToAddRightR[i] = nil
		linesToAddRightG[i] = nil
		linesToAddRightB[i] = nil
	end
end

function CowTip_LuaText:GameTooltip_GetUnit(this)
	if not lastUnitName then
		return self.hooks[GameTooltip].GetUnit(GameTooltip)
	end
	return lastUnitName, lastUnit
end

function CowTip_LuaText:OnTooltipSetUnit()
	local unit = CowTip:GetMouseoverUnit() or "mouseover"
	if UnitExists(unit) then
		self:OnTooltipSetUnit2()
	elseif not self:HasTimer("CowTip_LuaText_SetLater") then
		self:AddTimer("CowTip_LuaText_SetLater", 0.1, "OnTooltipSetUnit2")
	end
end

function CowTip_LuaText:OnTooltipSetUnit2()
	lastUnitName, lastUnit = self.hooks[GameTooltip].GetUnit(GameTooltip)
	local num = 2
	local text_2 = leftGameTooltipStrings[2]:GetText()
	if not text_2 then
		num = num - 1
	elseif not text_2:find(LEVEL_start) then
		-- has a guild
		num = num + 1
	end
	local mouseover_unit = CowTip:GetMouseoverUnit()
	if not UnitPlayerControlled(mouseover_unit) and not UnitIsPlayer(mouseover_unit) then
		-- npc, faction line
		local factionText = leftGameTooltipStrings[num+1]:GetText()
		if factionText == PVP then
			factionText = nil
		end
		if factionText and (factionList[factionText] or UnitFactionGroup(mouseover_unit)) then
			num = num + 1
		end
	end
	if not UnitIsConnected(mouseover_unit) or not UnitIsVisible(mouseover_unit) or UnitIsPVP(mouseover_unit) then
		-- offline, zone, or pvp
		num = num + 1
	end
	scanTT(mouseover_unit)
	self:RefixLines(num+1)
	self:AddRepeatingTimer("CowTip_LuaText_UpdateTexts", 1, "UpdateTexts")
end

function CowTip_LuaText:UpdateTexts()
	if not lastUnit or not UnitExists(mouseover_unit) then
		return
	end
	self:RefixLines(lastNumRegistered+1)
	
	GameTooltip:Show()
end

function CowTip_LuaText:OnTooltipHide()
	for i = 1, lastNumRegistered do
--		DogTag:RemoveFontString(leftGameTooltipStrings[i])
--		DogTag:RemoveFontString(rightGameTooltipStrings[i])
	end
	lastNumRegistered = 0
	lastUnitName, lastUnit = nil, nil
	
	self:RemoveTimer("CowTip_LuaText_UpdateTexts")
end

local function CleanCode(text)
--	text = DogTag:CleanCode(text)
	text = text:gsub("%s%s*", " ")
	return text
end

local function fixColor()
	GameTooltipTextLeft1:SetTextColor(1, 1, 1)
end

function CowTip_LuaText:PostTooltipSetUnit()
	self:AddTimer(0, fixColor)
end
